//
//  MainView.swift
//  nixar-ios-kickstart
//
//  Created by Muhammet Yıldız on 26.05.2022.
//

import SwiftUI
import nixar_api_ios

struct MainView: View {
  @State private var disableInvitationButton = true
  @State private var message = ""
  @State private var nixarAgent = Optional<NixarAgent>.none
  
  var body: some View {
    VStack(alignment: .center) {
      Button("Create Nixar Agent") {
        self.nixarAgent = NixarHelper.createAgent({message in
          print("Nixar Message:", message)
        })
        
        if let _ = nixarAgent {
          displayMessage("agent created");
          self.disableInvitationButton = false;
        } else {
          self.disableInvitationButton = true;
        }
      }.padding()
      
      Button("Create Invitation"){
        
        do{
          let invitation = try nixarAgent.self?.createLocalInvitation(
            label: "MyInvitation",
            endpoint: "http://example.com",
            seed: nil);
          displayMessage("\(try JSONHelper.encode(invitation))");
        } catch {
          print("error")
          print("Error text \(error)")
          displayMessage("\(error)")
        }
      }
      .disabled(self.disableInvitationButton)
      .padding()
      Button("DidCommAndSignVerifyTest"){
       NixarHelper.agentSignVerify({message in
            print("Nixar Message:", message)
          })
      }
      .padding()
      .padding()
      .padding()
      Text(self.message)
    }.onAppear {
      NixarHelper.initLogging();
      displayMessage("Initialized logging")
    }
  }
  
  func displayMessage(_ string: String) {
    DispatchQueue.main.asyncAfter(deadline: .now() + 0.5, execute: {
      withAnimation {
        self.message = string;
        print(self.message);
      }
    })
  }
}

struct MainView_Previews: PreviewProvider {
  static var previews: some View {
    MainView()
  }
}
