//
//  NixarHelper.swift
//  nixar-ios-kickstart
//
//  Created by Muhammet Yıldız on 27.05.2022.
//

import Foundation
import nixar_api_ios
import BigInt

class NixarHelper{
  
  static func initLogging() {
    NixarAgentFactory.initLogging(NixarAgentFactory.LOG_TRACE, {level, logMessage in
      print("Log", level, logMessage)
    });
  }
  
  static func createAgent(_ resultCallback: (String) -> ()) -> Optional<NixarAgent> {
    do{
      let genesisPath = try createGenesisFile();
      let agentInitParams = AgentInitParams(alias: "MyCoolIOSAgent",
                                            role: nil,
                                            seed: "0000000000000000000MYIOSCLIENT01",
                                            genesisPath: genesisPath)
      let passworCallback = { () in
        return "12345"
      }
      
      let agent = try NixarAgentFactory.createAgent(agentInitParams: agentInitParams, passwordCallback: passworCallback);
      
      resultCallback("Agent created successfully");
      
      return Optional.some(agent)
    } catch {
      print("error")
      print("Error text \(error)")
      resultCallback("\(error)")
      return Optional<NixarAgent>.none
    }
  }

  static func issueVerify(_ resultCallback: (String) -> ()) {
    //Plese don't use this function. It doesn't compile for now. Its here only for coding purposes. It will work in the next release
    do{
        let genesisPath = try createGenesisFile();
        let agentInitParams = AgentInitParams(alias: "MyCoolIOSIssuer",
        role: nil,
        seed: "0000000000000000000MYIOSISSUER01",
        genesisPath: genesisPath)
        let passworCallback = { () in
        return "12345"
        }

        let issuer = try NixarAgentFactory.createAgent(agentInitParams: agentInitParams, passwordCallback: passworCallback);
        let prover = try NixarAgentFactory.createAgent(agentInitParams: agentInitParams, passwordCallback: passworCallback);

        //TODO

        //region tag::GetTail[]
        //TODO
        // endregion end::GetTail[]

        //region tag::StoreTail[]
        //TODO
        // endregion end::StoreTail[]


        //region tag::CreateSchema[]
        let schemaName = "swift_schema"
        let attributeSet = ["name", "surname", "age", "gender"]
        let schemaVersion = "2.0"

        let schemaId = try issuer.createSchema(schemaName: schemaName, attributeSet: attributeSet, schemaVersion: schemaVersion)
        // endregion end::CreateSchema[]

        //region tag::CreateCredentialDefinition[]
        let credDefId = try issuer.createCredentialDefinition(schemaId: schemaId, isRevocable: true, issuanceType: .ISSUANCE_ON_DEMAND, maxCredNum: 20)
        // endregion end::CreateCredentialDefinition[]

        //region tag::CreateCredentialOffer[]
        let issuerNonce = BigUInt.randomInteger(withExactWidth: 80)
        // endregion end::CreateCredentialOffer[]

        let offer = try issuer.createCredentialOffer(credDefId: credDefId, issuerNonce:issuerNonce);

        //region tag::CreateCredentialRequest[]
        let request = try prover.createCredentialRequest(credOffer: offer)
        // endregion end::CreateCredentialRequest[]

        //region tag::CreateCredential[]
        var credValues = [String: AttributeValues]()
        credValues["name"] = AttributeValues(raw: "Mehmet");
        credValues["surname"] = AttributeValues(raw: "Öztürk");
        credValues["age"] = AttributeValues(raw: "30");
        credValues["gender"] = AttributeValues(raw: "male");

        let credential = try issuer.createCredential(credRequest: request, credValues: JSONHelper.encode(credValues), issuerNonce:issuerNonce)
        // endregion end::CreateCredential[]


        //region tag::StoreCredential[]
        let stored = try prover.storeCredential(credId: "myCredential", credential: credential.credential)
        // endregion end::StoreCredential[]

        //region tag::CreatePresentationRequest[]
        pres_req = '{"name": "IdentityPR", "version": "2.0", "nonce": "59240876", "requestedAttributes": {"attribute1_referent": {"name": "name"}, "attribute2_referent": {"name": "surname"}, "attribute3_referent": {"name": "gender"}, "attribute4_referent": {"name": "hobies"}}, "requestedPredicates": {"predicate1_referent": {"name": "age", "p_type": ">", "p_value": 10}}}'
        // endregion end::CreatePresentationRequest[]

        //region tag::CreatePresentation[]
        //if self attested value exist
        let selfAttestedValues = {}
        selfAttestedValues["attribute4_referent"] = "chess"
        var presentation = try prover.createPresentation(presentationRequest: presReq, selfAttestedValues: selfAttestedValues)
        // endregion end::CreatePresentation[]

        // region tag::VerifyPresentation[]

        //verifier verifies the presentation
        var result = try verifier.verifyPresentation(presentationRequest: presReq, presentation: presentation)
        print("Verified", result)
        XCTAssert(result == true)
        // endregion end::VerifyPresentation[]

        // region tag::RevokeCredential[]
        try issuer.revokeCredential(credRevIndex: credential.credRevId!, credDefId: credDefId)
        // endregion end::RevokeCredential[]
        } catch {
        print("error")
        print("Error text \(error)")
        resultCallback("\(error)")
        return Optional<NixarAgent>.none
    }
  }
  
  static func agentSignVerify(_ resultCallback: (String) -> ()){
    do{
      let genesisPath = try createGenesisFile();
      
      let passworCallback = { () in
        return "123456"
      }
      

      // region tag::CreateAgent[]
      var agentInitParams = AgentInitParams(alias: "TUBITAK DID IOS ISSUER",
                                            role: "ENDORSER",
                                            seed: "0000000TUBITAK DID IOS ISSUER 01",
                                            genesisPath: genesisPath)
      

      let agenta = try NixarAgentFactory.createAgent(agentInitParams: agentInitParams, passwordCallback: passworCallback);
      //can try-catch and check if agent creation failed

      // endregion end::CreateAgent[]

      resultCallback("AgentA created successfully");
      
      agentInitParams = AgentInitParams(alias: "TUBITAK_DID_IOS_PROVER",
                                            role: nil,
                                            seed: "00000000TUBITAK DID IOS PROVER 01",
                                            genesisPath: genesisPath)
      
      let agentb = try NixarAgentFactory.createAgent(agentInitParams: agentInitParams, passwordCallback: passworCallback);
      
      resultCallback("AgentB created successfully");

      // region tag::GetInvitation[]
      let publicInvitation = try agenta.getPublicInvitation()
      // endregion end::GetInvitation[]

      var connection2 = try findConnection(agent: agentb, alias: publicInvitation.label)
      
      if connection2 == nil {
        // region tag::ConnectionCreateRequest[]
        //Connect Agent A and Agent B
        //The seed is optional (may be nil). But please make sure that it is 32 ASCII characters (No Unicode please)
        let connectionSeed = "0ConnSEED on Agent B for Agent A"
        let encryptedConnectionRequest = try agentb.createConnectionRequest(invitation: publicInvitation, seed: connectionSeed)
        // endregion end::ConnectionCreateRequest[]

        // region tag::ConnectionAcceptRequest[]
        let encryptedConnectionResponse = try agenta.acceptConnectionRequest(requestMessage: encryptedConnectionRequest, alias: "agentb")
        // endregion end::ConnectionAcceptRequest[]

        // region tag::ConnectionAcceptResponse[]
        try agentb.acceptConnectionResponse(responseMessage: encryptedConnectionResponse)
        // endregion end::ConnectionAcceptResponse[]
        
        connection2 = try findConnection(agent: agentb, alias: publicInvitation.label)!
      }
      
      let connection = connection2!
      
      // region tag::ProverSign[]
      let signedData = "Some Data To Sign".data(using: .utf8)!.bytes
      print("Data To Sign \(signedData)")
      let base64EncodedSignature = try agentb.signWithDid(did: connection.myDid, data: signedData)
      print("Signature \(base64EncodedSignature)")
      // endregion end::ProverSign[]
      
      // region tag::ProverEncryptMessage[]
      //send the signature and the data to issuer (via didcomm)
      var packedSignature : [String: String] = [:]
      packedSignature["signedData"] = Data(signedData).base64EncodedString()
      packedSignature["signature"] = base64EncodedSignature;
      
      let messageString = try JSONHelper.encode(packedSignature);

      let encryptedMessage = try agentb.encryptMessage(fromDid: connection.myDid, toDid: connection.theirDid,
                                                   messageType: SwiftNixarMessageType.MessageTypeSimple, content: messageString);

      // endregion end::ProverEncryptMessage[]
      

      // issuer gets the data
      // region tag::IssuerDecryptMessageForVerification[]
      let decryptedMessage = try agenta.decryptMessage(message: try JSONHelper.encode(encryptedMessage));

      let pPackedSignature = try JSONHelper.decode([String: String].self, from: decryptedMessage.content.data(using: .utf8)!);

      guard let pSignedDataBase64 = pPackedSignature["signedData"] else {return}
      guard let pSignatureBase64 = pPackedSignature["signature"] else {return}

      guard let pSignedData = Data(base64Encoded: pSignedDataBase64)?.bytes else {return}
      guard let pSignature = Data(base64Encoded: pSignatureBase64)?.bytes else {return}
      // endregion end::IssuerDecryptMessageForVerification[]

      print("Data To Be Signed: \(pSignedDataBase64)");
      print("Signature: \(pSignatureBase64)");

      // region tag::IssuerVerifySignature[]
      //1. verify with their did from connection.
      let b1 = try agenta.verifyWithTheirDid(theirDid:decryptedMessage.senderDid, data: pSignedData, signature: pSignature);
      print("Verification Result \(b1)");

      //2. OR verify with their PUBLIC Key

      let oTheirDidAsDidDoc = try agenta.getTheirDidAsDidDoc(theirDid: decryptedMessage.senderDid);

      let b2 = try agenta.verifyWithDidPublicKey(didPublicKeyBase58: (oTheirDidAsDidDoc.publicKeys?[0].base58EncodedPublicKey)!, data: pSignedData, signature: pSignature);
      print("Verification Result \(b2)");

      // endregion end::IssuerVerifySignature[]
      
    } catch {
      print("error")
      print("Error text \(error)")
      resultCallback("\(error)")
    }
  }
  
  
  private static func findConnection(agent: NixarAgent, alias: String) throws -> Connection? {
    try agent.getConnectionByAlias(alias: alias)
  }
  
  public static func createGenesisFile() throws -> String {
    let name = "genesis.txn"
    
    let fileURL = try FileManager.default
      .url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: true)
      .appendingPathComponent(name)
    
    print("URL: " + String(describing: fileURL))
    
    //create genesis file
    if FileManager.default.createFile(atPath: fileURL.path, contents: nil, attributes: nil) {
      print("\ndosya oluşturuldu----", fileURL)
      
      let semaphore = DispatchSemaphore(value: 0)
      URLSession.shared.dataTask(with: URL(string: "http://test.bcovrin.vonx.io/genesis")!) {(data, response, error) in
        let genesisStr = String(data: data!, encoding: .utf8)!
        print("Genesis ", genesisStr)
        do{try genesisStr.write(to: fileURL.absoluteURL, atomically: true, encoding: .utf8)}catch {
          print("Cannot write genesis")
        }
        semaphore.signal()
      }.resume()
      
      
      semaphore.wait();
    }
    
    
    var str = fileURL.absoluteString
    if (str.starts(with: "file://")){
      str.removeSubrange(str.startIndex ..< "file://".endIndex)
    }
    
    return str
  }
  
}
